import numpy as np
import os
import sys
import time
import argparse
import json
from PIL import Image
from matplotlib import pyplot as plt
from scipy import io
from sklearn.decomposition import PCA
import cv2

import torch
import torch.utils.data as data
import torch.optim as optim
from torch.autograd import Variable

sys.path.insert(0, '../')
from model_siam import *
from utils import *
from region_to_bbox import *
from visualization import *

sys.path.insert(0, '../mdnet_model')
from mdnet_model import *

module_path = os.path.abspath(os.path.join('../ssd'))
if module_path not in sys.path:
    sys.path.append(module_path)
if torch.cuda.is_available():
    torch.set_default_tensor_type('torch.cuda.FloatTensor')
sys.path.insert(0, '../ssd')
from ssd import build_ssd


def compute_success_overlap(gt_bb, result_bb):
    thresholds_overlap = np.arange(0, 1.05, 0.05)
    n_frame = len(gt_bb)
    success = np.zeros(len(thresholds_overlap))
    iou = np.zeros(n_frame)
    for i in range(n_frame):
        iou[i] = overlap_ratio(gt_bb[i], result_bb[i])
    for i in range(len(thresholds_overlap)):
        success[i] = np.float(sum(iou > thresholds_overlap[i])) / n_frame
    return success


def display_tracking(image, result_bb, response, response2):
    image_show = np.array(image)
    show_frame(image_show, result_bb, 1, response, response2)


def scale_estimation(template, image, feature_net, \
                     target_sz, target_center):
    scales = [1, 0.96, 1.04]
    scores = np.zeros(len(scales))
    for j in range(len(scales)):
        scale = scales[j]
        scaled_target_sz = target_sz * scale
        context_ = context * (scaled_target_sz[0] + scaled_target_sz[1])
        scaled_target_sz_ = np.sqrt((scaled_target_sz[0] + context_) * (scaled_target_sz[1] + context_))
        # target_szz = target_sz
        # window_sz_ = float(search_sz) / exemplar_sz * scaled_target_sz_
        scaled_window_sz = float(search_sz) / exemplar_sz * scaled_target_sz_
        # scaled_window_sz = get_search_size(scaled_target_sz)
        # t1=time.time()
        patch = get_search_patch_siam(image, target_center, scaled_window_sz)
        patch = Variable(torch.from_numpy(patch[None, :]))
        patch = bilinear_upsample(patch, (search_sz, search_sz))
        if use_gpu:
            patch = patch.cuda()
        # t2 = time.time()
        # print('patch-resize --- %.4f' %(t2-t1))
        feat = feature_net.forward(patch)
        # feat = resize_feat_cos_window(feat, cos_window)
        response = feature_net.siam(feat, template)
        scores[j] = response.max().data[0]
        if scores[0] < 1.5:
            return target_sz
    scales[1] = scales[1] * 0.97
    scales[2] = scales[2] * 0.97
    idx = scores.argmax()
    return 0.41 * target_sz + 0.59 * np.round(target_sz * scales[idx])


# resizing features as same size as cos_window
def resize_feat_cos_window(feat, cos_window):
    num_channels = feat.size(1)
    cos_window_sz = cos_window.size(3)
    feat_data = feat.data.clone().cpu().numpy()
    # t1 = time.time()
    feat = cv2.resize(feat_data[0].transpose(1, 2, 0).astype('float32'), dsize=(cos_window_sz, cos_window_sz))
    feat = feat.transpose(2, 0, 1)
    feat = Variable(torch.from_numpy(feat).float(), requires_grad=False)
    if use_gpu:
        feat = feat.cuda()
    # t2 = time.time() - t1
    # print('resize--%.4f' % t2)
    feat = feat * cos_window
    return feat


def set_optimizer(model, lr_base, w_decay):
    param_list = []
    for k, p in model.named_parameters():
        param_list.append({'params': [p], 'lr': lr_base, 'weight_decay': w_decay})
    optimizer = optim.Adam(param_list)
    return optimizer


max_object_sz = 127


def set_max_object_sz(seq_name):
    if seq_name == 'Jump':
        return 70
    elif seq_name == 'Skater2':
        return 70
    elif seq_name == 'Girl2':
        return 70
    else:
        return max_object_sz


def get_sequence(data_dir, seq_name, data_label):
    # generate config from a sequence name
    if data_label is 'OTB':
        img_dir = os.path.join(data_dir, seq_name, 'img')
        gt_path = os.path.join(data_dir, seq_name, 'groundtruth_rect.txt')
    else:
        img_dir = os.path.join(data_dir, seq_name, 'color')
        gt_path = os.path.join(data_dir, seq_name, 'groundtruth.txt')

    included_extenstions = ['jpg', 'jpeg', 'png', 'bmp', 'gif']
    img_list = [fn for fn in os.listdir(img_dir)
                if any(fn.endswith(ext) for ext in included_extenstions)]
    # file_names.sort()
    # img_num = len(file_names)
    # img_list = os.listdir(img_dir)
    img_list.sort()
    img_list = [os.path.join(img_dir, x) for x in img_list]
    with open(gt_path) as f:
        gt_ = np.loadtxt((x.replace(',', ' ') for x in f))
    gt = np.zeros([len(gt_),4])
    for iii in range(len(gt_)):
        gt[iii,:] = region_to_bbox(gt_[iii,:])

    n_frames = len(img_list)
    if not n_frames == len(gt):
        img_list = img_list[0:len(gt)]
        n_frames = len(gt)
    init_bbox = gt[0]
    return img_list, init_bbox, gt


def ssd_detection(image_file, ssd_net, class_detection, class_number):
    image = cv2.imread(image_file)
    rgb_image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    x = cv2.resize(image, (300, 300)).astype(np.float32)
    x -= (104.0, 117.0, 123.0)
    x = x.astype(np.float32)
    x = x[:, :, ::-1].copy()
    x = torch.from_numpy(x).permute(2, 0, 1)
    xx = Variable(x.unsqueeze(0))  # wrap tensor in Variable
    if torch.cuda.is_available():
        xx = xx.cuda()
    y = ssd_net(xx)
    detections = y.data
    scale = torch.Tensor(rgb_image.shape[1::-1]).repeat(2)
    detection_bbox = []
    if class_detection:
        i = class_number
        j = 0
        # for j in range(detections.size(2)):
        while detections[0, i, j, 0] > 0.4:
            score = detections[0, i, j, 0]
            pt = (detections[0, i, j, 1:] * scale).cpu().numpy()
            detection_bbox_ = np.array([pt[0], pt[1], pt[2] - pt[0] + 1, pt[3] - pt[1] + 1, score], dtype=np.float32)
            detection_bbox.append(detection_bbox_)
            j += 1
    else:
        for i in range(detections.size(1)):
            # for j in range(detections.size(2)):
            j = 0
            while detections[0, i, j, 0] >= 0.4:
                score = detections[0, i, j, 0]
                pt = (detections[0, i, j, 1:] * scale).cpu().numpy()
                detection_bbox_ = np.array([pt[0], pt[1], pt[2] - pt[0] + 1, pt[3] - pt[1] + 1, score], dtype=np.float32)
                detection_bbox.append(detection_bbox_)
                j += 1

    # if not len(detection_bbox)==0:
    #     detection_bbox = np.stack(detection_bbox, axis=0)
    #     detection_bbox = np.sort(detection_bbox, axis=-1)
    return detection_bbox


def ssd_detection_first(image_file, gt, ssd_net):
    image = cv2.imread(image_file)
    rgb_image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    x = cv2.resize(image, (300, 300)).astype(np.float32)
    x -= (104.0, 117.0, 123.0)
    x = x.astype(np.float32)
    x = x[:, :, ::-1].copy()
    x = torch.from_numpy(x).permute(2, 0, 1)
    xx = Variable(x.unsqueeze(0))  # wrap tensor in Variable
    if torch.cuda.is_available():
        xx = xx.cuda()
    y = ssd_net(xx)
    detections = y.data
    scale = torch.Tensor(rgb_image.shape[1::-1]).repeat(2)
    iou_max = 0
    iou = 0
    class_detection = 0
    class_number = 0
    for i in range(detections.size(1)):
        j = 0
        while detections[0, i, j, 0] > 0.:
            score = detections[0, i, j, 0]
            pt = (detections[0, i, j, 1:] * scale).cpu().numpy()
            detection_bbox_ = np.array([pt[0], pt[1], pt[2] - pt[0] + 1, pt[3] - pt[1] + 1], dtype=np.float32)
            iou = _compute_iou(detection_bbox_, gt)
            if iou>0.8:
                class_detection = 1
                class_number = i
                break
            else:
                if iou>iou_max:
                    iou_max = iou
                    class_number = i
            j+=1

    if iou > 0.4:
        class_detection = 1

    return class_detection, class_number


def run_meta_crest(seq, ssd_net, display=False):

    '''load images and ground truth'''
    img_list, init_bbox, gt = get_sequence(data_dir, seq, 'VOT')
    max_object_sz = set_max_object_sz(seq)

    target_bbox1 = np.array(init_bbox)
    result = np.zeros((len(img_list), 4))
    result[0] = target_bbox1

    image = Image.open(img_list[0]).convert('RGB')

    image1 = Image.open(img_list[0]).convert('RGB')
    image_sz_resized = (int(image1.size[0]), int(image1.size[1]))
    image1_resized = image1.resize(image_sz_resized, Image.BILINEAR)

    # scale target bbox
    resized_target_bbox1 = np.round(target_bbox1)
    target_sz1 = resized_target_bbox1[2:]
    target_center1 = resized_target_bbox1[:2] + np.floor(target_sz1 / 2)
    # get cosine window(square shape)
    context_ = context * (target_sz1[0] + target_sz1[1])
    target_sz1_ = np.sqrt((target_sz1[0] + context_) * (target_sz1[1] + context_))
    window_sz_ = float(search_sz) / exemplar_sz * target_sz1_

    '''init mdnet model'''
    mdnet_model_ = mdnet_model(image1_resized, resized_target_bbox1)

    cos_window = np.outer(np.hanning(score_sz2), np.hanning(score_sz2))
    cos_window = cos_window[None, None, :, :].repeat(1, axis=1)
    cos_window = Variable(torch.from_numpy(cos_window).float(), requires_grad=False)
    if use_gpu:
        cos_window = cos_window.cuda()

    cos_window_p = np.outer(np.hanning(12), np.hanning(12))
    cos_window_p = cos_window_p[None, None, :, :].repeat(1, axis=1)
    cos_window_p = Variable(torch.from_numpy(cos_window_p).float(), requires_grad=False)
    if use_gpu:
        cos_window_p = cos_window_p.cuda()

    '''initialize the siamese fc network'''
    feature_net = FeatureExtractor_siam(feat_extractor_path)
    if use_gpu:
        feature_net = feature_net.cuda()
    feature_net.eval()

    '''utilize the first frame as template'''
    patch0 = get_search_patch_siam(image1_resized, target_center1, target_sz1_)
    patch0 = Variable(torch.from_numpy(patch0[None, :]))
    patch0 = bilinear_upsample(patch0, [exemplar_sz, exemplar_sz])
    if use_gpu:
        patch0 = patch0.cuda()
    feat0 = feature_net.forward(patch0)
    template = feat0

    '''detect the object and decide the class of tracked object'''
    class_detection, class_number = ssd_detection_first(img_list[0], target_bbox1, ssd_net)
    print(class_detection)

    '''parameters'''
    target_cell_sz1 = np.ceil(target_sz1_ / cell_sz)
    n_frame = len(img_list)
    target_center = target_center1
    target_sz = target_sz1
    resized_target_bbox = resized_target_bbox1
    window_sz_min = window_sz_ * 0.2
    window_sz_max = window_sz_ * 5.
    target_sz_min = target_sz1_ * 0.2
    target_sz_max = target_sz1_ * 5.
    scales = [1, 0.962, 1.04]
    time1 = time.time()
    train_frame = 0
    img_train_all = []
    label_train_all = []
    frame_train_all = []

    if display:
        dpi = 80.0
        figsize = (image.size[0] / dpi, image.size[1] / dpi)

        fig = plt.figure(frameon=False, figsize=figsize, dpi=dpi)
        ax = plt.Axes(fig, [0., 0., 1., 1.])
        ax.set_axis_off()
        fig.add_axes(ax)
        # im = ax.imshow(image, aspect='normal')
        im = ax.imshow(image)
        rect = plt.Rectangle(tuple(gt[0, :2]), gt[0, 2], gt[0, 3],
                             linewidth=3, edgecolor="#ff0000", zorder=1, fill=False)
        ax.add_patch(rect)

        plt.pause(.01)
        plt.draw()

    '''start tracking'''
    for i in range(n_frame - 1):
        if i > 0:
            resized_target_bbox_keep = resized_target_bbox.copy()
        else:
            resized_target_bbox_keep = resized_target_bbox1.copy()
        frame_num = i + 1

        '''siamenseFC tracking'''
        image = Image.open(img_list[frame_num]).convert('RGB')
        image_resized = image.resize(image_sz_resized, Image.BILINEAR)
        patch_ = get_search_patch_scale(image_resized, target_center, window_sz_)
        feat = feature_net.forward(patch_)
        response_ = feature_net.siam(feat, template)
        response_[1, :, :] = response_[1, :, :] * 0.97
        response_[2, :, :] = response_[2, :, :] * 0.97
        response_d = response_.data.cpu().numpy()
        new_scale_id = np.argmax(np.amax(response_d, axis=(1, 2)))
        window_sz_ = window_sz_ * scales[new_scale_id]
        if window_sz_ < window_sz_min:
            # target_sz_ = target_sz_min
            window_sz_ = window_sz_min
            new_scale_id = 0
        if window_sz_ > window_sz_max:
            # target_sz_ = target_sz_max
            window_sz_ = window_sz_max
            new_scale_id = 0

        response = response_[new_scale_id][None, None, :, :]
        response = bilinear_upsample(response, [score_sz2, score_sz2])
        score = response.max().data[0]
        if i == 0:
            score_init = score
        response.data = response.data - response.data.min()
        response.data = response.data / response.data.max()
        response.data = (1 - 0.25) * response.data + 0.25 * cos_window.data

        response_np = response.data[0][0].cpu().numpy()
        delta = np.argwhere(response_np.max() == response_np)[0] + 1  # zero-index -> original-index
        delta = delta - np.ceil((score_sz2) / 2.)
        delta_xcrop = delta * cell_sz * score_sz / score_sz2
        delta_in_frame = delta_xcrop * int(window_sz_) / search_sz
        target_center[0] = target_center[0] + delta_in_frame[1]
        target_center[1] = target_center[1] + delta_in_frame[0]
        target_center = valid_pos(target_center, image_sz_resized)
        target_sz = 0.41 * target_sz + 0.59 * target_sz * scales[new_scale_id]
        resized_target_bbox_keep = np.zeros(4)
        resized_target_bbox_keep[:2] = target_center - target_sz / 2
        resized_target_bbox_keep[2:] = target_sz


        '''evaluate the tracking score'''
        mdnet_score = mdnet_model_.forward(image_resized, resized_target_bbox_keep)
        if i == 0:
            init_mdnet_score = mdnet_score[0,1]

        '''tracking or detection'''
        if mdnet_score[0, 1] < 0.:
            if train_frame not in frame_train_all and len(frame_train_all)>0:
                mdnet_model_.re_train(img_train_all, label_train_all)
                train_frame = frame_train_all[-1]
                if train_frame==1:
                    train_frame=0
            detection_bbox = ssd_detection(img_list[frame_num], ssd_net, class_detection, class_number)

            '''get the top10 bounding box and use mdnet to choose the best one'''
            top10_bbox = np.zeros([5, 5], np.float32)
            iii = 0
            for jj in range(len(detection_bbox)):
                mdnet_score = mdnet_model_.forward(image_resized, detection_bbox[jj][0:4])
                re_bbox = np.copy(detection_bbox[jj][0:4])
                re_sz = re_bbox[2:]
                if (mdnet_score[0, 1] > init_mdnet_score/4.) and (re_sz.max() < target_sz_max):
                    print('detection')
                    # bbox_md, mdnet_score = mdnet_model_.re_detection(image_resized, detection_bbox[jj][0:4])
                    # if mdnet_score[0,1] > mdnet_score_init/4.:
                    #     resized_target_bbox_keep = np.copy(bbox_md)
                    #     resized_target_bbox = np.copy(bbox_md)
                    # else:
                    resized_target_bbox_keep = np.copy(detection_bbox[jj][0:4])
                    resized_target_bbox = np.copy(detection_bbox[jj][0:4])
                    target_sz = resized_target_bbox_keep[2:]
                    target_center = resized_target_bbox_keep[:2] + target_sz / 2
                    break
                    # if iii==5:
                    #     idx_max = np.argmax(top10_bbox[:,4])
                    #     print('detection')
                    #     resized_target_bbox_keep = np.copy(detection_bbox[idx_max][0:4])
                    #     target_sz = resized_target_bbox_keep[2:]
                    #     target_center = resized_target_bbox_keep[:2] + target_sz / 2
                    #     break
                    # else:
                    #     top10_bbox[iii, 0:4] = np.copy(detection_bbox[jj][0:4])
                    #     top10_bbox[iii, 4] = mdnet_score[0, 1]
                    #     iii = iii+1
        else:
            if mdnet_score[0, 1] > 0.:#init_mdnet_score/4.:
                print('tracking')
                resized_target_bbox = np.copy(resized_target_bbox_keep)
                # target_sz = 0.41 * target_sz + 0.59 * resized_target_bbox_keep[2:]
                target_sz = resized_target_bbox_keep[2:]
                target_center = resized_target_bbox_keep[:2] + target_sz / 2
                mdnet_model_.sample_generator.set_trans_f(0.6)


        result[frame_num] = np.round(resized_target_bbox)
        context_ = context * (target_sz[0] + target_sz[1])
        target_sz_ = np.sqrt((target_sz[0] + context_) * (target_sz[1] + context_))
        window_sz_ = float(search_sz) / exemplar_sz * target_sz_

        '''save training samples'''
        if mdnet_score[0, 1] > init_mdnet_score / 3.:
            patchn = get_search_patch_siam(image_resized, target_center, target_sz_)
            patchn = Variable(torch.from_numpy(patchn[None, :]))
            patchn = bilinear_upsample(patchn, [exemplar_sz, exemplar_sz])
            if use_gpu:
                patchn = patchn.cuda()
            featn = feature_net.forward(patchn)
            # template.data = (1 - 0.01) * template.data + 0.01 * featn.data
            image_train = image_resized
            bbox_train = resized_target_bbox
            if len(img_train_all) < 5:
                img_train_all.append(image_train)
                label_train_all.append(bbox_train)
                frame_train_all.append(frame_num)
            elif (frame_num - frame_train_all[-1]) > 5:
                del img_train_all[1]
                del label_train_all[1]
                del frame_train_all[1]
                img_train_all.append(image_train)
                label_train_all.append(bbox_train)
                frame_train_all.append(frame_num)

        if display:
            print('mdnet_score:%.2f--frame-num:%d' % (mdnet_score[0, 1], frame_num))
            # try:
            #     display_tracking(image, result[frame_num], response_np,
            #                      np.squeeze(response2.data.cpu().numpy()))  # np.squeeze(response_2.data.cpu().numpy())
            # except:
            #     display_tracking(image, result[frame_num], response_np,
            #                      response_np)  # np.squeeze(response_2.data.cpu().numpy())

            im.set_data(image)

            rect.set_xy(result[frame_num, :2])
            rect.set_width(result[frame_num, 2])
            rect.set_height(result[frame_num, 3])

            if display:
                plt.pause(.01)
                plt.draw()


                # overlap = overlap_ratio(gt[frame_num],result[frame_num])
                # print('%s(%d/%d) - score(%.3f), overlap(%.3f)'%(seq,frame_num,n_frame,score,overlap))

                # if (i+1)%update_interval==0:
                #     if len(label_update) >= update_interval:
                #         tracker_net.train()
                #         weight_end = model_update(weight_end, torch.stack(motion_map_l),online_optimizer_initcf1, online_optimizer_initcf2,
                #                                   torch.stack(label_update_m), tracker_net, online_optimizer, criterion, torch.stack(feat_update),
                #                                   torch.stack(label_update))

    time2 = time.time()
    times_ = time2 - time1
    return result, gt, n_frame, times_


def _compile_results(gt, bboxes, dist_threshold):
    l = np.size(bboxes, 0)
    gt4 = np.zeros((l, 4))
    new_distances = np.zeros(l)
    new_ious = np.zeros(l)
    n_thresholds = 50
    precisions_ths = np.zeros(n_thresholds)

    for i in range(l):
        # if i==114:
        #     a=1
        if np.isnan(gt[i,:].sum()):
            gt[i, :] = np.zeros([1,4],dtype=np.float32)
        gt4[i, :] = region_to_bbox(gt[i, :], center=False)
        # print(i)
        new_distances[i] = _compute_distance(bboxes[i, :], gt4[i, :])
        new_ious[i] = _compute_iou(bboxes[i, :], gt4[i, :])

    # what's the percentage of frame in which center displacement is inferior to given threshold? (OTB metric)
    precision = np.float(sum(new_distances < dist_threshold)) / np.size(new_distances) * 100

    # find above result for many thresholds, then report the AUC
    thresholds = np.linspace(0, 25, n_thresholds + 1)
    thresholds = thresholds[-n_thresholds:]
    # reverse it so that higher values of precision goes at the beginning
    thresholds = thresholds[::-1]
    for i in range(n_thresholds):
        precisions_ths[i] = np.float(sum(new_distances < thresholds[i])) / np.size(new_distances)

    # integrate over the thresholds
    precision_auc = np.trapz(precisions_ths)

    # per frame averaged intersection over union (OTB metric)
    iou = np.mean(new_ious) * 100

    return l, precision, precision_auc, iou


def _init_video(img_path, video):
    if 'vot' in img_path:
        video_folder = os.path.join(img_path, video)
    else:
        video_folder = os.path.join(img_path, video, 'img')
    frame_name_list = [f for f in os.listdir(video_folder) if f.endswith(".jpg")]
    frame_name_list = [os.path.join(video_folder, '') + s for s in frame_name_list]
    frame_name_list.sort()
    # with Image.open(frame_name_list[0]) as img:
    if video == 'David':
        frame_name_list = frame_name_list[299:]
    if video == 'Tiger1':
        frame_name_list = frame_name_list[5:]
    img = Image.open(frame_name_list[0])
    frame_sz = np.asarray(img.size)
    frame_sz[1], frame_sz[0] = frame_sz[0], frame_sz[1]

    # read the initialization from ground truth
    if 'vot' in img_path:
        gt_file = os.path.join(video_folder, 'groundtruth.txt')
    else:
        gt_name = video + '_gt.txt'
        gt_file = os.path.join(os.path.join(img_path, video, gt_name))
    gt_file = os.path.join(img_path, video, 'groundtruth_rect.txt')
    gt = np.genfromtxt(gt_file, delimiter=',')
    if gt.shape.__len__() == 1:  # isnan(gt[0])
        gt = np.loadtxt(gt_file)
    n_frames = len(frame_name_list)
    if not n_frames == len(gt):
        frame_name_list = frame_name_list[0:len(gt)]
        n_frames = len(gt)
    # assert n_frames == len(gt), 'Number of frames and number of GT lines should be equal.'

    return gt, frame_name_list, frame_sz, n_frames


def _compute_distance(boxA, boxB):
    a = np.array((boxA[0] + boxA[2] / 2, boxA[1] + boxA[3] / 2))
    b = np.array((boxB[0] + boxB[2] / 2, boxB[1] + boxB[3] / 2))
    dist = np.linalg.norm(a - b)

    assert dist >= 0
    assert dist != float('Inf')

    return dist


def _compute_iou(boxA, boxB):
    # determine the (x, y)-coordinates of the intersection rectangle
    xA = max(boxA[0], boxB[0])
    yA = max(boxA[1], boxB[1])
    xB = min(boxA[0] + boxA[2], boxB[0] + boxB[2])
    yB = min(boxA[1] + boxA[3], boxB[1] + boxB[3])

    if xA < xB and yA < yB:
        # compute the area of intersection rectangle
        interArea = (xB - xA) * (yB - yA)
        # compute the area of both the prediction and ground-truth
        # rectangles
        boxAArea = boxA[2] * boxA[3]
        boxBArea = boxB[2] * boxB[3]
        # compute the intersection over union by taking the intersection
        # area and dividing it by the sum of prediction + ground-truth
        # areas - the intersection area
        iou = interArea / float(boxAArea + boxBArea - interArea)
    else:
        iou = 0

    assert iou >= 0
    assert iou <= 1.01

    return np.float(iou)


use_gpu = True

## options
data_dir = '/home/lpx/Desktop/code-vot2018/VOT2018/sequences'

cf_num_channels = 32
vgg_num_channels = 512
output_sigma_factor = 0.1
motion_sigma_factor = 0.6
cell_sz = 4.0
exemplar_sz = 127
search_sz = 255
score_sz2 = 264
score_sz = 33
context = 0.5

online_lr_base = 0.000007
online_w_decay = 0
n_init_updates = 1
n_online_updates = 2
update_interval = 5
n_update_batch = 5
max_db_size = n_online_updates * n_update_batch
# cf1 parameters
init_cf1_num = 100
cf_lr = 7e-5
cf2_lr = 7e-6
init_com_weight = Variable(torch.from_numpy(np.array([1, 0.05, 1], dtype=np.float32)))

feat_extractor_path = '../models/baseline-conv5_e55.mat'
tracker = 'init%02d-lr%.6f' % (n_init_updates, online_lr_base)

score_thres = 0.0

np.random.seed(4)
torch.manual_seed(5)
torch.cuda.manual_seed(6)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--display', action='store_true')
    args = parser.parse_args()
    args.display = 1
    seqs = os.listdir(data_dir)
    seqs.sort()

    '''initialize ssd model'''
    ssd_net = build_ssd('test', 300, 21)  # initialize SSD
    ssd_net.load_weights('../ssd/weights/ssd300_mAP_77.43_v2.pth')

    n_seq = len(seqs)
    speed = np.zeros(n_seq)
    precisions = np.zeros(n_seq)
    precisions_auc = np.zeros(n_seq)
    ious = np.zeros(n_seq)
    lengths = np.zeros(n_seq)
    # n_seq = 1
    action_ = []
    for i in range(7,n_seq):
        # i=31
        idx = i
        seq = seqs[i]

        result_bb, gt, n_frame, time_ = run_meta_crest(seq, ssd_net, display=args.display)
        speed[idx] = (n_frame - 1) / time_  # basketball2.5
        lengths[idx], precisions[idx], precisions_auc[idx], ious[idx] = _compile_results(gt, result_bb,
                                                                                         20)
        print(str(i) + ' -- ' + seq + \
              ' -- Precision: ' + "%.2f" % precisions[idx] + \
              ' -- Precisions AUC: ' + "%.2f" % precisions_auc[idx] + \
              ' -- IOU: ' + "%.2f" % ious[idx] + \
              ' -- Speed: ' + "%.2f" % speed[idx] + ' --')
        # print
        # np.savetxt("/home/lpx/Desktop/code/results/basecom/%s_basecom.txt" % seq,
        #                       np.round(result_bb), delimiter=',')
    # l = len(action_)
    # aa = np.zeros([l, 3])
    # for i in range(l):
    #     a = action_[i]
    #     a1 = (a == 1).sum()
    #     a2 = (a == 2).sum()
    #     a3 = (a == 3).sum()
    #     aal = a1 + a2 + a3
    #     aa[i, 0] = a1 / aal
    #     aa[i, 1] = a2 / aal
    #     aa[i, 2] = a3 / aal
    tot_frames = np.sum(lengths)
    mean_precision = np.mean(precisions)
    mean_precision_auc = np.mean(precisions_auc)
    mean_iou = np.mean(ious)
    mean_speed = np.mean(speed)
    print('-- Overall stats (averaged per frame) on ' + str(1) + ' videos (' + str(tot_frames) + ' frames) --')
    print(' -- Precision ' + "(%d px)" % 20 + ': ' + "%.2f" % mean_precision + \
          ' -- Precisions AUC: ' + "%.2f" % mean_precision_auc + \
          ' -- IOU: ' + "%.2f" % mean_iou + \
          ' -- Speed: ' + "%.2f" % mean_speed + ' --')

    for i in range(n_seq):
        print(seqs[i])
    for i in range(n_seq):
        print(round(precisions[i], 2))
    print(round(mean_precision, 2))

    for i in range(n_seq):
        print(round(ious[i], 2))
    print(round(mean_iou, 2))

    # res = {}
    # res['results'] = []
    # res['results'].append({'res': result_bb.round().tolist(), 'type': 'rect',
    #                      'len': len(result_bb), 'anno':gt.round().tolist()})
    # json.dump(res, open('../result/otb/' + seq + '_' + tracker + '.json', 'w'), indent=2)
    # io.savemat('../result/otb/' + seq +  '_' + tracker + '.mat',res)

